// File: alsa_plugin.cpp
// Author: Neil Macvicar (blackmogu@vfemail.net)
// Date: 14/12/2002
// Abstract: Implementation of the ALSA Mixer plugin for QTVision
// Revision Control:
// 0.1: Initial Implementation

/*
 * Copyright (C) 2002 Neil Macvicar <blackmogu@vfemail.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "qtvision_alsa.h"
#include <iostream>
#include <kdebug.h>
#include <stdio.h>
#include <qwidget.h>
#include <qframe.h>
#include <qlayout.h>
#include <qcombobox.h>
#include <qlabel.h>
#include <klocale.h>
#include <kgenericfactory.h>
#include <kconfig.h>

//-----------------------------------------------------------------------------
// Constructors and destructors
//-----------------------------------------------------------------------------

QTVisionALSA::QTVisionALSA(QObject *parent, const char *name, const QStringList &/*args*/) : QVMixerPlugin("alsamixer", parent, name)
{
  const char *sig="[QTVisionALSA()]";

  kdDebug() << PLUGIN_ID << sig << " initialising plugin" << endl;
  
  card_selector = NULL;
  element_selector = NULL;
  active_HCTL.truncate(0);
  active_mixer = NULL;
  active_element = NULL;

  QTVisionALSA::loadConfig();
  // we need to check if this configuration is still available

  kdDebug() << PLUGIN_ID << sig << " plugin initialised" << endl;
}

QTVisionALSA::~QTVisionALSA()
{
  const char *sig="[~QTVisionALSA()]";

  kdDebug() << PLUGIN_ID << sig << " unloading plugin" << endl;

  if(active_mixer != NULL && !active_HCTL.isEmpty())
    QTVisionALSA::detachMixer(active_mixer, active_HCTL);
  
  _cfg->setGroup("ALSA Mixer");
  _cfg->writeEntry("Mute Setting", muted_flag);
  
  kdDebug() << PLUGIN_ID << sig << " plugin unloaded" << endl;
}

//-----------------------------------------------------------------------------
// Configuration member functions
//-----------------------------------------------------------------------------

void QTVisionALSA::saveConfig(void)
{
  const char *sig="[saveConfig()]";

  QMap<int, QString>::iterator it;
  QString mixer_element = element_selector->currentText();
  int retval;
  char *card_name;

  kdDebug() << PLUGIN_ID << sig << "saving configuration" << endl;

  // this is called when OK is clicked in the config. Find out the HCTL ID 
  // from the human readable card name held in card_selector
  for(it = probed_cards.begin() ; it != probed_cards.end() ; it++) {
    snd_card_get_name(it.key(), &card_name); 
    if(card_selector->currentText() != card_name) continue;
    break;
  }

  retval = useCardMixerElement(it.data(), mixer_element);
  
  if(retval != 0) return; // don't bother saving this config, it's invalid
  
  _cfg->setGroup("ALSA Mixer");
  _cfg->writeEntry("HCTL ID", active_HCTL);
  _cfg->writeEntry("Mixer Element", element_selector->currentText());
  _cfg->sync();

  kdDebug() << PLUGIN_ID << sig << "configuration saved" << endl;
}

void QTVisionALSA::loadConfig(void)
{
  const char *sig="[loadConfig()]";
  QString HCTL_id, mixer_element; 
  
  kdDebug() << PLUGIN_ID << sig << " loading configuration" << endl;

  _cfg->setGroup("ALSA Mixer");
  HCTL_id = _cfg->readEntry("HCTL ID", "");
  mixer_element = _cfg->readEntry("Mixer Element", "");
  
  if(QTVisionALSA::useCardMixerElement(HCTL_id, mixer_element) != 0) {
    kdDebug() << PLUGIN_ID << sig << " failed loading" << endl;
    return;
  }
  
  // unmute the mixer device if it is indeed muted
  if(QTVisionALSA::setMuted(false) != 0) {
    kdDebug() << PLUGIN_ID << sig << " failed to unmute mixer element" << endl;
    return;
  }

  kdDebug() << PLUGIN_ID << sig << " configuration loaded" << endl;
  
  return;
}

//-----------------------------------------------------------------------------
// Atomic member functions
//-----------------------------------------------------------------------------

snd_mixer_t *QTVisionALSA::attachMixer(QString &HCTL_id)
{
  const char *sig="[attachMixer()]";
  snd_mixer_t *tmp_mixer;
  int retcode;
  
  kdDebug() << PLUGIN_ID << sig << " attaching a mixer to "
	    << HCTL_id << endl;

  if((retcode = snd_mixer_open(&tmp_mixer, 0)) != 0) {
    kdDebug() << PLUGIN_ID << sig << " snd_mixer_open() error "
	      << retcode << endl;
    return(NULL);
  }

  if((retcode = snd_mixer_attach(tmp_mixer, HCTL_id)) != 0) {
    kdDebug() << PLUGIN_ID << sig << " snd_mixer_attach() error "
	      << retcode << endl;
    snd_mixer_close(tmp_mixer);
    return(NULL);
  }

  snd_mixer_selem_register(tmp_mixer, NULL, NULL);

  kdDebug() << PLUGIN_ID << sig << " mixer attach successful" << endl;

  return(tmp_mixer);
}

int QTVisionALSA::detachMixer(snd_mixer_t *mixer, const char *HCTL_id)
{ 
  const char *sig="[detachMixer()]";
  int retval;

  kdDebug() << PLUGIN_ID << sig << " detaching mixer from " << HCTL_id << endl;

  if((retval = snd_mixer_detach(mixer, HCTL_id)) != 0) {
    kdDebug() << PLUGIN_ID << sig << " snd_mixer_detach() failed, "
	      << "error " << retval << endl;
    return(retval);
  }
  if((retval = snd_mixer_close(mixer)) != 0) {
    kdDebug() << PLUGIN_ID << sig << " snd_mixer_close() failed, "
	      << "error " << retval << endl;
    return(retval);
  }
  
  kdDebug() << PLUGIN_ID << sig << " detached successfully" << endl;
 
  return 0;
}

int QTVisionALSA::loadMixerElements(snd_mixer_t *mixer)
{
  const char *sig="[loadMixerElements()]";
  snd_mixer_elem_t *mixer_element;
  snd_mixer_selem_id_t *sid;
  int retval, element_count = 0;

  kdDebug() << PLUGIN_ID << sig << " discovering mixer elements" << endl;
 
  // no API documentation for this yet
  snd_mixer_selem_id_alloca(&sid);

  // the data in mixer_elements is no longer pertinent. Erase
  mixer_elements.clear();
 
  // load the mixer elements
  if((retval = snd_mixer_load(mixer)) != 0) {
    kdDebug() << PLUGIN_ID << sig << " snd_mixer_load() " 
	      << "failed, error " << retval << endl; 
    return(retval);
  }
   
  for(mixer_element = snd_mixer_first_elem(mixer) ; mixer_element ; 
      mixer_element = snd_mixer_elem_next(mixer_element)) {
    element_count++;
    snd_mixer_selem_get_id(mixer_element, sid);

    // filter out any non-playback mixer elements
    if(!snd_mixer_selem_is_active(mixer_element)) continue;
    if(!snd_mixer_selem_has_playback_volume(mixer_element)) continue;
    if(!snd_mixer_selem_has_playback_switch(mixer_element)) continue;

    mixer_elements.insert(mixer_element, snd_mixer_selem_id_get_name(sid));
    kdDebug() << PLUGIN_ID << sig << " + " 
	      << snd_mixer_selem_id_get_name(sid) << endl;
  }

  kdDebug() << PLUGIN_ID << sig << " elements discovered : " << element_count 
	    << endl;
  kdDebug() << PLUGIN_ID << sig << " playback elements : " 
	    << mixer_elements.count() << endl;

  return(0);
}

int QTVisionALSA::probeDevices(void)
{
  const char *sig="[probeDevices()]";
  int card_id = -1;
  int retval;

  kdDebug() << PLUGIN_ID << sig << " querying ALSA driver" << endl;

  // clear out our probed_cards QMap
  probed_cards.clear();

  // Repopulate our probed_cards map
  while((retval = snd_card_next(&card_id)) == 0) {
    if(card_id == -1) break;
    
    // store the card information
    probed_cards.insert(card_id, QString("hw:%1").arg(card_id));
  }
  
  // check to see if we exited the loop due to an error
  if(retval != 0) {
    kdDebug() << PLUGIN_ID << sig << " snd_card_next() failed, code " 
	      << retval << endl;
  }

  kdDebug() << PLUGIN_ID << sig << " cards discovered : " 
	    << probed_cards.size() << endl;
  
  return(0);
}

//-----------------------------------------------------------------------------
// Compound member functions
//-----------------------------------------------------------------------------

int QTVisionALSA::useCardMixerElement(QString &HCTL_id, QString &mixer_element)
{
  const char *sig="[useCardMixerElement()]";
  QMap<snd_mixer_elem_t *, QString>::iterator it;

  kdDebug() << PLUGIN_ID << sig << " request " << HCTL_id << '(' 
	    <<	mixer_element << ')' << endl;

  // detach and unload any existing mixer 
  if(!active_HCTL.isEmpty() && active_mixer != NULL) {
    if(QTVisionALSA::detachMixer(active_mixer, active_HCTL) != 0)
      return(1);
    active_HCTL.truncate(0);
    active_mixer = NULL;
  }
  
  // attach a mixer to the HCTL interface
  if((active_mixer = QTVisionALSA::attachMixer(HCTL_id)) == NULL)
    return(1);
  active_HCTL = HCTL_id;
  
  // load all of the available elements for this card
  if(QTVisionALSA::loadMixerElements(active_mixer) != 0) {
    detachMixer(active_mixer, active_HCTL);
    active_HCTL.truncate(0);
    active_mixer = NULL;
    return(1);
  }
  
  // obtain the active element for this card
  for(it = mixer_elements.begin() ; it != mixer_elements.end() ; it++) {
    if(mixer_element != it.data()) continue;
    break;
  }
  
  active_element = it.key();
  
  kdDebug() << PLUGIN_ID << sig << " request successful" << endl;
  return(0);
}

QWidget *QTVisionALSA::configWidget(QWidget *parent, const char *name)
{
  const char *sig="[configWidget()]";
  QFrame *config_frame = new QFrame(parent, name);
  QGridLayout *config_grid = new QGridLayout(config_frame, 7, 7);
  
  // we need two pull-down lists, one for sound cards, and one for mixer
  // elements for the selected card
  QLabel *Label1 = new QLabel(i18n("Available cards:"), config_frame);
  QLabel *Label2 = new QLabel(i18n("Mixer elements:"), config_frame);

  kdDebug() << PLUGIN_ID << sig << " creating configuration screen" << endl;

  // Initialise all pertinent objects.  
  card_selector = new QComboBox(config_frame, "Card List");
  element_selector = new QComboBox(config_frame, "Element List");

  mixer_elements.clear();

  config_grid->addMultiCellWidget(Label1, 0, 0, 0, 2);
  config_grid->addMultiCellWidget(Label2, 1, 1, 0, 2);
  config_grid->addMultiCellWidget(card_selector, 0, 0, 3, 7);
  config_grid->addMultiCellWidget(element_selector, 1, 1, 3, 7);

  // scan for available cards
  QTVisionALSA::probeDevices();
   
  // populate the sound card QComboBox with available cards
  if(!probed_cards.empty()) {
    QMap<int, QString>::iterator it;
    char *card_name;

    for(it = probed_cards.begin() ; it != probed_cards.end() ; it++) {
      snd_card_get_name(it.key(), &card_name);
      card_selector->insertItem(card_name);
    }
  }

  // discover what mixer elements are available for the first entry in the
  // card_selector QComboBox and add them to the element_selector QComboBox
  if(!probed_cards.empty()) {
    snd_mixer_t *mixer;
    QMap<snd_mixer_elem_t *, QString>::iterator it;

    mixer = QTVisionALSA::attachMixer(probed_cards.begin().data());
    QTVisionALSA::loadMixerElements(mixer);
    QTVisionALSA::detachMixer(mixer,probed_cards.begin().data());
    
    for(it = mixer_elements.begin() ; it != mixer_elements.end() ; it++)
      element_selector->insertItem(it.data());
  }

  // attach a signal handler for the event of a new card being selected 
  connect(card_selector, SIGNAL(activated(const QString &)), 
	  this, SLOT(cardChanged(const QString &)));

  kdDebug() << PLUGIN_ID << sig << " configuration screen created" << endl;

  return(config_frame);
}

void QTVisionALSA::cardChanged(const QString &card_name)
{
  const char *sig="[cardChanged()]";
  snd_mixer_t *mixer;
  QMap<int, QString>::iterator it;

  kdDebug() << PLUGIN_ID << sig << " loading mixer elements for " 
	    << card_name << endl;

  // clear the element_selector list
  element_selector->clear();
  
  // obtain the HCTL ID for the given card name
  for(it = probed_cards.begin() ; it != probed_cards.end() ; it++) {
    char *tmp = NULL;

    if(snd_card_get_name(it.key(), &tmp) != 0)
      return;

    if(card_name != tmp) continue;
    break;
  }

  if((mixer = QTVisionALSA::attachMixer(it.data())) != NULL) {  
    QMap<snd_mixer_elem_t *, QString>::iterator it2;

    QTVisionALSA::loadMixerElements(mixer);
    for(it2 = mixer_elements.begin() ; it2 != mixer_elements.end() ; it2++) 
      element_selector->insertItem(it2.data());
    
    QTVisionALSA::detachMixer(mixer, it.data());
  }

  element_selector->setCurrentItem(0);

  kdDebug() << PLUGIN_ID << sig << " elements added to QComboBox" << endl;
  return;
}

int QTVisionALSA::setMuted(bool mute)
{
  const char *sig="[setMuted()]";
  snd_mixer_selem_channel_id_t ch_left = SND_MIXER_SCHN_FRONT_LEFT;
  int playback;

  kdDebug() << PLUGIN_ID << sig << " request active element mute=" 
	    << mute << endl;

 if(active_element == NULL){
    kdDebug() << PLUGIN_ID << sig << " There is no active element !" << endl;
    return(1);
  }

 snd_mixer_selem_get_playback_switch(active_element, ch_left, &playback);
 
 if((playback && mute) || (!playback && !mute))
   snd_mixer_selem_set_playback_switch_all(active_element, !playback);
 
 muted_flag = mute;

 kdDebug() << PLUGIN_ID << sig << ' ' << mixer_elements[active_element] 
	   << " on " << active_HCTL << " muted=" << mute << endl;
 return 0;
}

bool QTVisionALSA::muted(void)
{
  return(muted_flag);
}

int QTVisionALSA::setVolume(int LeftVolume, int RightVolume)
{
  if(!muted_flag) {
    volume_left = LeftVolume;
    volume_right = RightVolume;
  }

  return(0);
}

int QTVisionALSA::volumeLeft(void)
{
  //const char *sig="[volumeLeft()]";
  long min,max,vol;

  snd_mixer_selem_channel_id_t front_left = SND_MIXER_SCHN_FRONT_LEFT;

  if(active_mixer == NULL) return(0);
  snd_mixer_selem_get_playback_volume_range(active_element, &min, &max);

  // transform the volume into soundcard format
  vol = ((volume_left * (max - min)) / 100) + min;

  // set the volume level on the channel
  snd_mixer_selem_set_playback_volume(active_element, front_left, vol);
  
  //kdDebug() << PLUGIN_ID << sig << " left channel volume " << volume_left
  //	    << "% = step " << vol << endl;

  return(volume_left);
}

int QTVisionALSA::volumeRight(void)
{
  //const char *sig="[volumeRight()]";
  long min,max,vol;

  snd_mixer_selem_channel_id_t front_right = SND_MIXER_SCHN_FRONT_RIGHT;

  if(active_mixer == NULL) return(0);
  snd_mixer_selem_get_playback_volume_range(active_element, &min, &max);

  // transform the volume into soundcard format
  vol = ((volume_left * (max - min)) / 100) + min;

  // set the volume level on the channel
  snd_mixer_selem_set_playback_volume(active_element, front_right, vol);
  
  //kdDebug() << PLUGIN_ID << sig << " right channel volume " << volume_left
  //	    << "% = step " << vol << endl;

  return(volume_right); 
}


int QTVisionALSA::volume(void)
{
  const char *sig="[volume()]";

  kdDebug() << PLUGIN_ID << sig << "WHO CALLED THIS !?" << endl;
  // discover the volume level of a channel
  return(0);
}

int QTVisionALSA::setMixer(const QString &mixer)
{
  kdDebug() << "WHO CALLED setMixer with " << mixer << "!!!" << endl;
  return 0;
}

//-----------------------------------------------------------------------------

K_EXPORT_COMPONENT_FACTORY(qtvision_alsa, KGenericFactory<QTVisionALSA>);

#include "qtvision_alsa.moc"

